import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.ArrayList;
import javax.swing.*;
import javax.swing.event.*;

/**
 * Usability test of marker order.
 * <p>
 * @author sashhr
 */
public class Test
{           
   /** Number of markers per trial (distractors + target) */
   private static final int _nMarker = 5;
   
   /** Grid size. */ 
   private static final int _gridSize = 3;
   
   /** Number of trials per set. */
   private static final int _nTrial = 4 * _gridSize * _gridSize;
   
   /** @serial  True iff paused. */
   private static boolean _isPaused = false;
   
   /** @serial  True iff paused at start of data collection. */
   private static boolean _isData = false;
   
   /** @serial  Participant number. */
   private static int _user = 0;
   
   /** Marker names. */
   private static final String[] _markerNames = new String[]{
      "Circle      ",  
      "Plus        ", 
      "X           ",  
      "Triangle    ", 
      "Square      ",  
      "Asterisk    ",  
      "Diamond     ",  
      "IBeam       ", 
      "Union       "
   };

   /** Markers as bit Strings. */
   private static final String[] _markerStrings = new String[]{
      "0110010010100100110000000",  // Circle
      "0010000100111110010000100",  // Plus
      "1000101010001000101010001",  // X
      "0010001010010101000111111",  // Triangle
      "1111010010100101111000000",  // Square
      "0010010101011101010100100",  // Asterisk
      "0010001010100010101000100",  // Diamond
      "0111000100001000010001110",  // IBeam 
      "1000110001100011000101110"   // Union 
   };

   /** Practice markers as bit Strings. */
   private static final String[] _practiceStrings = new String[]{
      "0010001010100011000111111",  // Home
      "0101010101100010101000100",  // Heart
      "0100110101101011010110010",  // Tilde
      "1001010101101011010101001",  // Wreath
      "0100001111010101111000010",  // Weave
      "0000010001111111000100000",  // H
      "1111110001010100101000100",  // TriangleDown 
      "0001101101100010110100011",  // TriangleLeft
      "1100010110100011011011000"   // TriangleRight
   };

   /** Markers as integers. */
   private static int[] _markers = null;

   /** Number of sets = _markerNames.length - _nMarker + 1 */
   private static final int _nSet = _markerNames.length - _nMarker + 1;
   
   /** Number of pauses. */
   private static final int _nPause = _nSet - 1;
   
   /** Practice markers as integers. */
   private static int[] _practice = null;
   
   /** Marker detection test design. */
   private static int[] _markersDetTest = null;
   
   /** Marker detection target list. */
   private static int[] _markersDetTarget = null;
   
   /** Marker detection test X coordinates. */
   private static int[][] _markersDetX = null;
   
   /** Marker detection test Y coordinates. */
   private static int[][] _markersDetY = null;
   
   /** Y coordinate of prompt. */
   private static final int _yPrompt = 15;
   
   /** Initial prompt. */
   private static final String[] HELLO = new String[]{
      "Thank you for participating in usability tests of colors and markers.",
      "",
      "Please ensure that your monitor has 1024x768 resolution,",
      "24-bit color ('True Color'), and 75Hz refresh rate.  If you are",
      "not certain your monitor has these settings, click on the Exit",
      "button and check your settings, then start this program again.",
      "If you are unsure how to check or change your monitor settings,",
      "contact Cindy Huffman or Frank Wimmer.",
      "",
      "Click one of the buttons below to Start the test or to Exit."
   };
   
   /** Markers detection practice test prompt. */
   private static final String[] MARKERS_DET_PRACTICE = new String[]{
      "Please read these instructions before continuing.",
      "",
      "In this experiment, you will be shown displays of shapes.", 
      "On each trial, the display will contain a target shape",
      "that is different from all the other shapes.",
      "",
      "Press the '4', '5', '1', or '2' key to specify the quadrant",
      "of the display where you see the target shape:",
      "                                                 4  |  5   ",
      "                                                ----|----",
      "                                                 1  |  2   ",
      "",
      "The first set of trials are practice trials.",
      "",
      "Press the spacebar to begin the practice trials."
   };
   
   /** Markers detection data collection prompt. */
   private static final String[] MARKERS_DET_DATA = new String[]{
      "You have finished the practice trials.",
      "",
      "If you have any question about the experiment, then please",
      "click the Exit button and contact Cindy Huffman or Frank Wimmer.",
      "Do not continue if you are unsure what you are supposed to do.",
      "",
      "The data from the remaining trials will be used in our analysis.",
      "Speed and accuracy are important. Respond as quickly", 
      "as possible, but not so quickly that you make mistakes.",
      "",
      "Click the Start button to begin the data collection trials."
   };
   
   /** Markers simulation data collection done. */
   private static final String[] MARKERS_DET_PAUSE = new String[]{
      "Time to take a break.  Thanks for your help in this study.",
      "",
      "When you are ready to start again, ",
      "click the Start button to resume."
   };
   
   /** Markers detection data collection done. */
   private static final String[] MARKERS_DET_DONE = new String[]{
      "You have finished data collection.  Thanks for your participation.",
      "",
      "Please copy the file ",
      "to the folder \\\\d4582\\Public\\marker_detection_results.",
      "",
      "Click the Exit button to end this program."
   };
        
   /** Test prompt, first line. */
   private static final String PROMPT0 = "Target:    ";
   
   /** Test prompt, second line. */
   private static final String PROMPT1 = "Press the '4', '5', '1', or '2' key to specify the quadrant.";
   
   /** Initial prompt. */
   private static final String INITIAL = "Click one of the buttons below to start a test or to exit.";
   
   /** Space prompt. */
   private static final String SPACE = "Press the space bar to continue.";
   
   /** Wait prompt. */
   private static final String WAIT = "Please wait...";
   
   /** Error prefix. */
   private static final String ERROR = "ERROR: ";
   
   /** Start button name. */
   private static final String START = "Start";
        
   /** Exit button name. */
   private static final String EXIT = "Exit";
   
   /** Start button. */
   private static JToggleButton _buttonStart = new JToggleButton( START );
        
   /** Exit button. */
   private static JToggleButton _buttonExit = new JToggleButton( EXIT );
        
   /**
    * Return number of items in string separated by spaces.
    * <p>
    * @param   s  string
    * @return  number of items in string separated by spaces
    */
   private static int parseSize( String s )
   {  int size = 0;
        int slength = s.length();
      if( !s.trim().equals( "" ))
         size++;
      int i = 0;
      for( ; ( s.charAt( i ) == ' ' ) && ( i < slength ); )
          i++;
      for( ; ( i < slength ); i++ )
      {  if( s.charAt( i ) == ' ' )
            size++;
      }
      return( size );
   }

   /**
    * Parse strings separated by spaces.
    * <p>
    * @param   s  string
    * @return  a  array of strings
    */
   public static String[] parseStrings( String s )
   {  String[] a = new String[ parseSize( s )];
      int slength = s.length();
      int alength = a.length;
      int prev = 0;
      int next = 0;
      int i = 0;
      while( prev < slength )
      {  for( ; ( prev < slength ) && ( s.charAt( prev ) == ' ' ); )
             prev++;
          next = prev;
          for( ; ( next < slength ) && ( s.charAt( next ) != ' ' ); )
             next++;
         String sub = s.substring( prev, ( next > 0 ) ? next : slength );
         prev = next + 1;
         if( i < alength )
         {  a[ i ] = sub.trim();
            i++;
         }
      }
      return( a );
   }
   
   /**
    * Parse integers separated by spaces.
    * <p>
    * @param   s  string
    * @return  a  array of ints
    */
   public static int[] parseInts( String s )
   {  int[] a = new int[ parseSize( s )];
      int slength = s.length();
      int alength = a.length;
      int prev = 0;
      int next = 0;
      int i = 0;
      while( prev < slength )
      {  for( ; ( prev < slength ) && ( s.charAt( prev ) == ' ' ); )
             prev++;
          next = prev;
          for( ; ( next < slength ) && ( s.charAt( next ) != ' ' ); )
             next++;
         String sub = s.substring( prev, ( next > 0 ) ? next : slength );
         prev = next + 1;
         if( i < alength )
         {  a[ i ] = Integer.parseInt( sub.trim());
            i++;
         }
      }
      return( a );
   }
   
   /**
    * Returns a random order of specified length.
    * <p>
    * @param   order  order array
    * @return  randomized order array
    */
   public static int[] randomOrder( int[] order )
   {  for( int i = 0; ( i < order.length ); i++ )
         order[ i ] = -1;
      for( int i = order.length; ( i > 0 ); i-- )
      {  int j = ( int )( i * Math.random());
         int k = 0;
         for( ; ( order[ k ] >= 0 ); )
            k++;
         for( int l = 0; ( l < j ); l++ )
         {  k++;
            for( ; ( order[ k ] >= 0 ); )
               k++;
         }
         order[ k ] = i - 1;
         // order[ i - 1 ] = i - 1;  for debugging
      }
      return( order );
   }

   /**
    * Usability test of colors and marker shapes.
    */
   public static void main( String[] args )
   {  
      /**
       * Filler panel.
       */
      class FillPanel extends JPanel
      {  
         /**
          * Constructor.
          * <p>
          * @param  width  width
          * @param  height  height
          */
         public FillPanel( int width, int height )
         {  super();
            setPreferredSize( new Dimension( width, height ));
         }
      }
      
      /**
       * Test display panel.
       */
      class DisplayPanel extends JPanel 
         implements KeyListener, ActionListener, ChangeListener
      {
         /** @serial  User action. */
         String _action = "";
        
         /** @serial  Current index. */
         int _index = 0;
   
         /** @serial  Current subtrial. */
         int _subtrial = 0;
   
         /** @serial  True iff in wait state. */
         boolean _isWaiting = false;
   
         /** @serial  True iff doing practice tests. */
         boolean _isPractice = false;
           
         /** @serial  Test results. */
         ArrayList _results = null;
         
         /** @serial  Start time. */
         long _time0 = 0L;
         
         /** @serial  End time. */
         long _time1 = 0L;
         
         /** @serial  Target marker. */
         int _target = 0;
         
         /** @serial  Target marker X coordinate. */
         int _x0 = 0;
         
         /** @serial  Target marker Y coordinate. */
         int _y0 = 0;
         
         /** @serial  Quadrant. */
         char _quadrant = '4';
         
         /**
          * Constructor.
          */
         public DisplayPanel()
         {  
            // Set size and background.
            super();
            setPreferredSize( new Dimension( 600, 400 ));
            setBackground( Color.white );
        
            // Add listeners.
            addKeyListener( this );
            _buttonStart.addActionListener( this );
            _buttonExit.addActionListener( this );
         }
       
         /**
          * Draw the specified shape.
          * <p>
          * @param    graphics  Graphics context
          * @param    color     Marker color
          * @param    shape     Marker shape
          * @param    x0        horizontal center
          * @param    y0        vertical center
          */
         public void drawShape( Graphics g, Color color, int shape, int x0, int y0 )
         {  g.setColor( color );
            int mult = 1;
            int x = x0 - 2;
            int y = y0 - 2;
            for( int r = 4; ( r >= 0 ); r-- )
            {  for( int c = 4; ( c >= 0 ); c-- )
               {  if( shape % ( mult * 2 ) >= mult )
                     g.drawLine( x + c, y + r, x + c, y + r );
                  mult *= 2;
               }
            }
         }
     
         /**
          * Draw the specified prompt in the test panel.
          * <p>
          * @param graphics    Graphics context
          * @param prompt      Prompt
          * @param distance    Distance from top of display; <0 => center of display
          * @param isCleared   True iff background cleared before drawing
          */
         public void drawPrompt( Graphics g, String prompt, int distance, boolean isCleared )
         {  int width = getSize().width;
            int height = getSize().height;
            int x0 = width / 2;
            int y0 = height / 2;
            if( isCleared )
            {  g.setColor( getBackground());
               g.fillRect( 0, 0, width, height );
            }
            g.setColor( Color.black );
            g.drawString( prompt, 
               x0 - g.getFontMetrics().stringWidth( prompt ) / 2, 
               (( distance < 0 ) ? y0 + 5 : _yPrompt + 5 + 16 * distance ));
         }
     
         /**
          * Draw the specified prompt at the center of the test panel.
          * <p>
          * @param graphics    Graphics context
          * @param prompt      Prompt
          */
         public void drawPrompts( Graphics g, String[] prompt )
         {  g.setColor( getBackground());
            g.fillRect( 0, 0, getSize().width, getSize().height );
            int lineHeight = g.getFontMetrics().getHeight();
            int yOff = ( getSize().height - lineHeight * prompt.length ) / 2 +
               g.getFontMetrics().getAscent();
            int lineWidth = 0;
            for( int i = 0; ( i < prompt.length ); i++ )
            {  if( lineWidth < g.getFontMetrics().stringWidth( prompt[ i ]))
                  lineWidth = g.getFontMetrics().stringWidth( prompt[ i ]);
            }
            int xOff = ( getSize().width - lineWidth ) / 2;
            g.setColor( Color.black );
            for( int i = 0; ( i < prompt.length ); i++ )
            {  g.drawString( prompt[ i ], xOff, yOff + lineHeight * i );
            }
         }
     
         /**
          * Generate a file name for the current test.
          */
         public String getFileName()
         {  String fileName = "MO";
            if( _nMarker < 10 )
               fileName += "0";
            fileName += _nMarker;
            if( _user < 100 )
               fileName += "0";
            if( _user < 10 )
               fileName += "0";
            fileName += _user + ".txt";
            return( fileName );
         }
     
         /**
          * Format an integer in a specified field width.
          * <p>
          * @param  i      Integer value
          * @param  width  Field width
          */
         public String format( int i, int width )
         {  String s = "";
            int mult = 10;
            for( int j = 1; ( j < width ); j++ )
            {  if( i < mult )
                  s += " ";
               mult *= 10;
            }
            s += i;
            return( s );
         }
     
         /**
          * Generate a file name for the current test.
          * <p>
          * @param  c  Color
          */
         public String getColorString( Color c )
         {  String s = "";
            s += format( c.getRed()  , 3 ) + " ";
            s += format( c.getGreen(), 3 ) + " ";
            s += format( c.getBlue() , 3 );
            return( s );
         }
        
         /**
          * Paint this component.
          * <p>
          * @param  g     Graphics context
          */
         public void paint( Graphics g )
         {  super.paint( g );
            
            // Draw the initial instructions...
            if( _action.equals( "" ))
               drawPrompts( g, HELLO );
            
            // ...or hack around the refresh problem.  
            // Probably all the drawing code should move here.
            else if( _isPractice && _isWaiting && ( _index == 0 ))
            {  drawPrompts( g, MARKERS_DET_PRACTICE );
            }
         }
     
         /**
          * Marker detection test.
          * <p>
          * @param  g     Graphics context
          * @param  c     Key character
          */
         public void runMarkerDetection( Graphics g, char c ) 
         {
            // If paused, display prompt and return.
            if( _isPaused )
            {  if( c == '0' )
               {  _isPaused = false;
                  _isData = false;
                  drawPrompt( g, SPACE, -1, true );
                  _buttonStart.setSelected( true );
               }
               else
                  drawPrompts( g, _isData ? MARKERS_DET_DATA : MARKERS_DET_PAUSE );
               return;
            }
               
            // Get panel dimensions and center.
            int width = getSize().width;
            int height = getSize().height;
            int x0 = width / 2;
            int y0 = height / 2;
            
            // Draw the instruction screen for practice tests...
            if( _isPractice && ( _index < 0 ))
            {  _index = 0;
               _isWaiting = true;
               drawPrompts( g, MARKERS_DET_PRACTICE );
            }
            
            // ...or read and store the user's input.
            else if( _index < _markersDetTest.length )
            {  if( _isWaiting )
               {  
                  // Display the trial...
                  if( c == ' ' )
                  {
                     // Initialize.
                     _isWaiting = false;
                     _time0 = System.currentTimeMillis();
                     g.setColor( Color.white );
                     g.fillRect( 0, 0, width, height );
                     
                     // Get index, target marker, and set.
                     // Deriving target from participant ID supports Latin Square design.
                     int index = _markersDetTest[ 
                        _isPractice ? _markersDetTest.length - 1 - _index : _index ];
                     _subtrial = index % _nTrial;
                     _target = _nMarker - 1 + _markersDetTarget[ index ];
                     int[] markers = _isPractice ? _practice : _markers;
                                          
                     // Draw target marker.
                     _x0 = x0 + 5 + _markersDetX[ index ][ 0 ];
                     _y0 = y0 + 5 + _markersDetY[ index ][ 0 ];
                     _quadrant = ( _y0 < y0 ) ? (( _x0 < x0 ) ? '4' : '5' ) 
                                              : (( _x0 < x0 ) ? '1' : '2' );
                     if( _x0 < x0 )
                        _x0 -= 5;
                     else
                        _x0 += 5;
                     if( _y0 < y0 )
                        _y0 -= 5;
                     else
                        _y0 += 5;
                     drawShape( g, Color.black, markers[ _target ], _x0, _y0 );
                     
                     // Draw distractor markers.
                     for( int n = 1; ( n < _nTrial ); n++ )
                     {  int x1 = x0 + 5 + _markersDetX[ index ][ n ];
                        int y1 = y0 + 5 + _markersDetY[ index ][ n ];
                        if( x1 < x0 )
                           x1 -= 5;
                        else
                           x1 += 5;
                        if( y1 < y0 )
                           y1 -= 5;
                        else
                           y1 += 5;
                        drawShape( g, Color.black, _markers[ n % ( _nMarker - 1 )], x1, y1 );
                     }
                     
                     // Draw prompt.
                     if( _isPractice )
                        drawPrompt( g, PROMPT1, 0, false );
                     drawPrompt( g, PROMPT0, 4, false );
                     drawShape( g, Color.black, markers[ _target ], 250, _yPrompt + 16 * 4 );
                  }
               }
               
               // ...or store the result of the user's input.
               else if(( c == '4' ) || ( c == '5' ) || ( c == '1' ) || ( c == '2' ))
               {  _isWaiting = true;
                  _time1 = System.currentTimeMillis();
                  if( !_isPractice )
                     _results.add( new String( format( _nMarker, 2 ) + "  " + format( _user, 3 ) + "  " + 
                        format( _index + 1, 3 ) + " " + 
                        format( _subtrial + 1, 3 ) + "  " +
                        format( _nMarker, 2 ) + "  " + format( _target - 1, 1 ) + "  " + 
                        _markerNames[ _target ] + "  " + 
                        format( _x0, 4 ) + "  " + format( _y0, 4 ) + "  " + 
                        _quadrant + "  " + c + "  " + (( _quadrant == c ) ? "1" : "0" ) + 
                        format(( int )( _time1 - _time0 ), 7 )));
//                  drawPrompt( g, WAIT, -1, true );
//                  long time = System.currentTimeMillis();
//                  for( ; ( time - _time1 < 2000 ); )
//                     time = System.currentTimeMillis();
                  drawPrompt( g, SPACE, -1, true );
                  _index++;
               }
            }
            
            // When done, write the results to the output file...
            if( _index == _markersDetTest.length )
            {  String fileName = getFileName();
               String[] done = MARKERS_DET_DONE;
               done[ 2 ] += fileName;
               drawPrompts( g, done );
               try
               {  PrintWriter out = new PrintWriter(
                     new BufferedWriter( new FileWriter( fileName )));
                  Object[] results = _results.toArray();
                  for( int i = 0; ( i < results.length ); i++ )
                     out.println(( String )results[ i ]);
                  out.close();
               }
               catch( Exception e )
               {  System.out.println( "" + e );
               }
               _buttonStart.setSelected( false );
               _index++;
            }
            
            // ...otherwise if done with practice, display the data prompt...
            else if( _isPractice && ( _index >= _nTrial ))
            {  _index = 0;            
               _isPractice = false;
               _isPaused = true;       
               _isData = true;
               drawPrompts( g, MARKERS_DET_DATA );
               _buttonStart.setSelected( false );
            }
            
            // ...otherwise if time to pause, display the pause prompt.
            else if(( c != ' ' ) && 
               ( 0 < _index ) && ( _index < _markersDetTest.length ))
            {  int pauseIndex = _markersDetTest.length / ( _nPause + 1 );
               if( _markersDetTest.length % ( _nPause + 1 ) > 0 )
                  pauseIndex++;
               if( _index % pauseIndex == 0 )
               {  _isPaused = true;
                  drawPrompts( g, MARKERS_DET_PAUSE );
                  _buttonStart.setSelected( false );
               }
            }
         }
     
         /**
          * Keyboard events.
          */
         public void keyPressed( KeyEvent k ) {}
         public void keyReleased( KeyEvent k ) {}
         public void keyTyped( KeyEvent k ) 
         {  
            // Get graphics context and keyboard input.
            Graphics g = getGraphics();
            char c = k.getKeyChar();
              
            // Perform one iteration of the appropriate test.
            if( _action.equals( START ))
               runMarkerDetection( g, c );
            else if( _action.equals( EXIT ))
               System.exit( 0 );             
         }
         
         /**
          * Slider events.
          */
         public void stateChanged( ChangeEvent e )
         {  requestFocus();
         }
         
         /**
          * Command button actions.
          */
         public void actionPerformed( ActionEvent e ) 
         {  
            // Deselect all other toggle buttons.
            _action = e.getActionCommand();
            if( !_action.equals( START ))
               _buttonStart.setSelected( false );
            if( !_action.equals( EXIT ))
               _buttonExit.setSelected( false );
           
            // Perform the requested action.
            if( _action.equals( EXIT ))
               System.exit( 0 );   
            requestFocus();
            if( !_isPaused )
            {  _index = -1;
               _results = new ArrayList( 100 );
               _isPractice = true;
               _isWaiting = false;
            }
            keyTyped( new KeyEvent( this, 0, 0, 0, 0, '0', 0 ));          
         }
      }
      
      // Create the frame and the test panel.
      JFrame frame = new JFrame( "Markers Order Test" );
      frame.setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );
      JPanel testPanel = new JPanel( new GridBagLayout());
      GridBagConstraints c = new GridBagConstraints();
      c.fill = GridBagConstraints.BOTH;
      c.weightx = 1.0;
      c.weighty = 1.0;
      c.gridx = 0;
      c.gridy = 0;
      DisplayPanel displayPanel = new DisplayPanel();
      testPanel.add( displayPanel, c );
      c.fill = GridBagConstraints.HORIZONTAL;
      c.weighty = 0.0;
      c.gridy = 1;
      frame.getContentPane().add( testPanel, BorderLayout.CENTER );
      
      // Get the system version.  If less than 1.4, abort the test.
      double version = 0.0;
      try
      {  version = Double.parseDouble( System.getProperty( "java.version" ).substring( 0, 3 ));
      }
      catch( Exception ex )
      {  
      }
      if( version < 1.4 )
      {  JOptionPane.showMessageDialog( null, 
            "Java version is " + version + ".  Please install Java 1.4.", 
            "Test aborted", JOptionPane.ERROR_MESSAGE ); 
         System.exit( 0 );
      }
      
      // Prompt the user for their participant number.
      final JDialog prompt = new JDialog(( Frame )null,  
         "Please Enter Your Participant Number", true );
      prompt.setSize( new Dimension( 400, 100 ));
      prompt.setResizable( false );
      prompt.setLocationRelativeTo( null );
      prompt.getContentPane().setLayout( new GridBagLayout());
      c.fill = GridBagConstraints.NONE;
      c.anchor = GridBagConstraints.EAST;
      c.weightx = 1.0;
      c.weighty = 1.0;
      c.gridx = 0;
      c.gridy = 0;
      final JLabel label = new JLabel( "Participant Number:  " );
      prompt.getContentPane().add( label, c );
      c.anchor = GridBagConstraints.WEST;
      c.gridx = 1;
      final JTextField field = new JTextField();
      field.setPreferredSize( new Dimension( 100, 20 ));
      prompt.getContentPane().add( field, c );
      c.anchor = GridBagConstraints.CENTER;
      c.gridx = 0;
      c.gridy = 1;
      c.gridwidth = 2;
      final JButton ok = new JButton( "OK" );
      ok.setPreferredSize( new Dimension( 100, 20 )); 
      prompt.getContentPane().add( ok, c ); 
      prompt.getRootPane().setDefaultButton( ok ); 
      ok.setEnabled( false );
      field.getDocument().addDocumentListener( new DocumentListener()
      {  public void changedUpdate( DocumentEvent e ) 
         {  try
            {  _user = Integer.parseInt( field.getText().trim());
               ok.setEnabled( _user > 0 );
            }
            catch( Exception ex )
            {  ok.setEnabled( false );
            }
         }
         public void insertUpdate( DocumentEvent e ) 
         {  changedUpdate( e );
         }
         public void removeUpdate( DocumentEvent e ) 
         {  changedUpdate( e );
         }
      });
      ok.addActionListener( new ActionListener()
      {  public void actionPerformed( ActionEvent e ) 
         {  prompt.setVisible( false );
         }
      });
      prompt.setVisible( true );
      
      // Get the participant number.  If not valid, abort the test.
      try
      {  _user = Integer.parseInt( field.getText().trim());
      }
      catch( Exception ex )
      {  
      }
      if( _user <= 0 )
      {  JOptionPane.showMessageDialog( null, 
            "Participant number is invalid.  Please restart the test.", 
            "Test aborted", JOptionPane.ERROR_MESSAGE ); 
         System.exit( 0 );
      }
 
      // Convert the marker strings to integers.
      _markers = new int[ _markerStrings.length ];
      for( int i = 0; ( i < _markerStrings.length ); i++ )
      {  _markers[ i ] = Integer.parseInt( 
            _markerStrings[ i ], 2 );
      }
      _practice = new int[ _practiceStrings.length ];
      for( int i = 0; ( i < _practiceStrings.length ); i++ )
      {  _practice[ i ] = Integer.parseInt( 
            _practiceStrings[ i ], 2 );
      }
      
      // Generate target and distractor positions for each trial.
      // The target and distractors fill the grid.
      // They are distributed evenly and randomly within each quadrant.
      _markersDetX = new int[ _nSet * _nTrial ][];
      _markersDetY = new int[ _nSet * _nTrial ][];
      int[] x = new int[ _nTrial ];
      int[] y = new int[ _nTrial ];
      int nDistractor = _gridSize * _gridSize;
      int[] order = new int[ nDistractor ];
      for( int subTrial = 0; ( subTrial < _nTrial ); subTrial++ )
      {  
         // Generate random coordinates for symbols in each quadrant.
         // Store the target locations, evenly distributed across all possible locations.
         int target = -1;
         for( int quadrant = 0; ( quadrant < 4 ); quadrant++ )
         {  order = randomOrder( order );
            for( int distractor = 0; ( distractor < nDistractor ); distractor++ )
            {  int index = quadrant * nDistractor + distractor;
               x[ index ] = 10 * ( order[ distractor ] % _gridSize ); 
               y[ index ] = 10 * ( order[ distractor ] / _gridSize ); 
               if( quadrant % 2 == 0 )
                  x[ index ] -= 10 * _gridSize;
               if( quadrant < 2 )
                  y[ index ] -= 10 * _gridSize;
               if( subTrial == quadrant * nDistractor + order[ distractor ])
                  target = quadrant * nDistractor + distractor;
            }
         }

         // Swap coordinates so that the target is the first symbol.
         int xTemp = x[ 0 ];
         int yTemp = y[ 0 ];
         x[ 0 ] = x[ target ];
         y[ 0 ] = y[ target ];
         x[ target ] = xTemp;
         y[ target ] = yTemp;

         // Store the same coordinates for each set.
         for( int set = 0; ( set < _nSet ); set++ )
         {  int index = set * _nTrial + subTrial;
            _markersDetX[ index ] = new int[ _nTrial ];
            _markersDetY[ index ] = new int[ _nTrial ];
            System.arraycopy( x, 0, _markersDetX[ index ], 0, _nTrial );
            System.arraycopy( y, 0, _markersDetY[ index ], 0, _nTrial );
         }
      }
      
      // Store the target symbols.
      _markersDetTarget = new int[ _nSet * _nTrial ];
      for( int i = 0; ( i < _markersDetTarget.length ); i++ )
         _markersDetTarget[ i ] = i / _nTrial;

      // Randomize the trials.  Maintain the order of the initial trial for each set; 
      // these may be discarded to reduce variance due to the pause between sets.
      _markersDetTest = new int[ _nSet * _nTrial ];
      _markersDetTest = randomOrder( _markersDetTest );
      for( int i = 0; ( i < _markersDetTest.length ); i++ )
      {  if( _markersDetTest[ i ] % _nTrial == 0 )
         {  int j = _markersDetTest[ _markersDetTest[ i ]];
            _markersDetTest[ _markersDetTest[ i ]] = _markersDetTest[ i ];
            _markersDetTest[ i ] = j;
         }
      }
      
      // For each block, if two consecutive trials have targets in the same quadrant,
      // swap the second consecutive trial with a later trial to remove this effect.
      for( int i = 0; ( i < _nSet ); i++ )
      {  for( int j = 1; ( j < _nTrial ); j++ )
         {  if( getQuadrant( i * _nTrial + j - 1 ) == getQuadrant( i * _nTrial + j ))
            {  int k = j + 1;
               boolean equal = true;
               for( ; ( k < _nTrial ) && equal; k++ )
               {  equal = ( getQuadrant( i * _nTrial + j ) == getQuadrant( i * _nTrial + k ));
               }
               if( !equal )
               {  int temp = _markersDetTest[ i * _nTrial + j ];
                  _markersDetTest[ i * _nTrial + j ] = _markersDetTest[ i * _nTrial + k - 1 ];
                  _markersDetTest[ i * _nTrial + k - 1 ] = temp;
               }
            }
         }
      }
      
      /* For debugging, print all the coordinates.
      for( int i = 0; ( i < _markersDetTest.length ); i++ )   
      {  System.out.println( "$$$$$ " + i );
         for( int j = 0; ( j < _markersDetX[ 0 ].length ); j++ )
            System.out.println( "   " + _markersDetX[ i ][ j ] + "   " + 
               _markersDetY[ i ][ j ]);
      } */
      
      // Add the fill panels and the control panel.
      frame.getContentPane().add( new FillPanel( 280, 160 ), BorderLayout.NORTH );
      frame.getContentPane().add( new FillPanel( 280, 160 ), BorderLayout.EAST );
      frame.getContentPane().add( new FillPanel( 280, 160 ), BorderLayout.WEST );
      JPanel controlPanel = new JPanel( new GridBagLayout());
      c = new GridBagConstraints();
      c.fill = GridBagConstraints.BOTH;
      c.weightx = 1.0;
      c.weighty = 1.0;
      c.gridx = 0;
      c.gridy = 0;
      controlPanel.add( new FillPanel( 280, 160 ), c );
      JPanel buttonPanel = new JPanel();
      _buttonStart.setPreferredSize( new Dimension( 225, _buttonStart.getPreferredSize().height ));
      _buttonExit.setPreferredSize( new Dimension( 225, _buttonExit.getPreferredSize().height ));
      buttonPanel.add( _buttonStart, c );
      buttonPanel.add( _buttonExit, c );
      c.gridy = 1;
      controlPanel.add( buttonPanel, c );
      frame.getContentPane().add( controlPanel, BorderLayout.SOUTH );
      
      // Show the frame and let the user run the test.
      frame.setResizable( false );
      frame.pack();
      frame.setLocation( 0, 0 );
      frame.setSize( 1024, 718 );
      frame.setVisible( true );
   }
   
   /**
    * Return quadrant for specified trial.
    * <p>
    * @param   i  trial
    * @return  quadrant
    */
   public static int getQuadrant( int i )
   {  int j = _markersDetTest[ i ];
      int x = _markersDetX[ j ][ 0 ];
      int y = _markersDetY[ j ][ 0 ];
      int q = 0;
      if(( x >= 0 ) && ( y < 0 ))
         q = 1;
      else if(( x < 0 ) && ( y >= 0 ))
         q = 2;
      else if(( x >= 0 ) && ( y >= 0 ))
         q = 3;
      return( q );
   }
}
